#!/usr/bin/perl -w
#
# Author: Lulin Song 2005
#
# $Id: DriverGenerator.pl,v 1.23 2006/03/02 16:46:13 lsong Exp $
#
# To create a c driver for stand alone matlab program according description
# file
#
# Usage:
#
# DriverGenerator.pl descriptionFileName driverFileName libname
# eg.
#
# DriverGenerator.pl paramDescription.txt swg.c  libswg
#

if( $#ARGV < 2 )
{
   print <<EOF;
Usage: $0 descriptionFileName driverFileName sharedLibName
eg.
$0 paramDescription.txt swg.c libswg
EOF
   exit(0);
}

($inputFile,$outputFile,$libName) = @ARGV;

if( $outputFile !~ /(\.c$){1}/ )
{
   print <<EOF;
Output file name must end with '.c' extension.
EOF
   exit(0);
}

open($PARAMFILE,$inputFile) or die "Can't open $inputFile: $!\n"; $idx=0;
@paramNos=();
@charNames=();
%dataTypes=();
%rowNums=();
%colNums=();
%descriptions=();
@exampleDesc=();

print <<EOF;
You have the following parameters in your description file.
-----------------------------------------------------------
EOF

LINE: while( <$PARAMFILE> ){
    next LINE if /^#/; # skip comment lines
    next LINE if /^$/; # skip blank lines
    # Get example lines
    if ($_ =~ /^\>/ )
    { my $eg = $_; 
        chop ($eg);
        $eg =~ tr/^\>//d; 
        push(@exampleDesc,$eg);
    }
    else
    {
      # get six fields from every line
      ($paramNo, $charName,$dataType,$rowNum,$colNum,@desc)=split(" ", $_);
      print "$paramNo, $charName, $dataType, $rowNum, $colNum, @desc\n";
      my $descs=();
      foreach (@desc)
      { $descs .= $_.' '; }
      $paramNos[$idx]=$paramNo;
      $charNames[$idx]=$charName;
      $dataTypes{$paramNo}= $dataType;
      $rowNums{$paramNo}= $rowNum;
      $colNums{$paramNo}= $colNum;
      $descriptions{$paramNo}=$descs;
    }
    $idx++;
};
print <<EOF;
-----------------------------------------------------------
EOF

close($PARAMFILE);

# Create comment block
open($OUTFILE,">$outputFile") or die "Can't open $outputFile: $!\n";
CreateCommentBlock($OUTFILE,$libName);
close($OUTFILE);
print "Comment block is created.\n";

# Create include block in .c file
open($OUTFILE,">>$outputFile") or die "Can't open $outputFile: $!\n";
CreateIncludeBlock($OUTFILE,$libName, \%dataTypes);
print "Include block is created.\n";

# Create input parameter option struct block in .c file
CreateInputParameterStructBlock($OUTFILE,$outputFile,\%dataTypes,\@paramNos,\@charNames,\%rowNums,\%colNums);
print "Input parameter struct block is created.\n";

# Create prototype block in .c file
CreatePrototypeBlock($OUTFILE,$outputFile,\%dataTypes,\@charNames,\@paramNos);
print "Prototype block is created.\n";

# Create main function block in .c file
CreateMainFunctionBlock($OUTFILE,$libName,$outputFile,\%dataTypes,\%rowNums,\%colNums, \@paramNos,\@charNames,\%descriptions);
print "Main function  block is created.\n";

# Create getParameters function in .c file
CreateGetCommandlineOptionBlock($OUTFILE,$outputFile,\%dataTypes,\%rowNums,\%colNums, \@paramNos,\@charNames,\%descriptions);
print "Main Command line option block is created.\n";

# Create printUsage function in .c file
CreatePrintUsageBlock($OUTFILE,$outputFile,\%dataTypes,\%rowNums,\%colNums, \@paramNos,\@charNames,\%descriptions,\@exampleDesc);
print "Print usage block is created.\n";

# Create writeOut function in .c file
CreateWriteOutDataBlock($OUTFILE,$outputFile,\%dataTypes,\%rowNums,\%colNums, \@paramNos,\@charNames);
print "Write out data block is created.\n";

close($OUTFILE);

print "Successfully created $outputFile .\n";

sub CreateCommentBlock{
my ($fh,$myLibName) = @_;
my $when = gmtime();
print $fh <<EOF;
/* Do not edit. Automatically generated by DriverGenerator.pl script.
*
* Author: Lulin Song (International Research Institute for Climate Society)
*
* Generated: $when (GMT)
*
* Driver program. Take parameters from command line and call shared
* library $myLibName.so created by Matlab compiler and write out the
* output data to a file.
*
*/
EOF
};

sub CreateIncludeBlock {
my ($fh,$myLibName, $myDataTypeRef) = @_;
my $includeStrings=" ";
foreach $myKey ( sort keys %$myDataTypeRef ){
    $myType = $myDataTypeRef->{$myKey};
  if( $myType =~ /^(ascii)$/ )
  { $includeStrings ="#include \"asciiRW.h\""; last;}
  if( $myType =~ /^(netcdf)$/ )
  { $includeStrings ="#include \"netcdfRW.h\""; last;}
}

print $fh <<EOF;
#include "$myLibName.h"

#include <stdio.h>
#include <matrix.h>
#include <stdarg.h>

$includeStrings

EOF
};

sub CreateInputParameterStructBlock {
my($fh, $myoutFileName,$myDataTypeRef, $myParamNoRef,$charNameRef,$rowNoRef,$colNoRef) = @_;
$myoutFileName =~ s/(\.c)$//;
my $structName = $myoutFileName."Opts";
my $typeString=" ";
foreach my $myNo ( @$myParamNoRef ){
    my $key = $charNameRef->[$myNo-1];
    if( $key !~ /^-$/ )
    {
      if( $key !~ /^(\+)$/ )
      {
        my $myType = $myDataTypeRef->{$myNo};
        SWITCH: {
        if($myType =~ /^(string)$/) 
          {$typeString .="char option".$myNo."[128];\n  "; last SWITCH;}
        if($myType =~ /^(int)$/) 
          {$typeString .="int option".$myNo.";\n  "; last SWITCH;}
        if($myType =~ /^(double)$/) 
          {$typeString .="double option".$myNo.";\n  "; last SWITCH;}
        if($myType =~ /^(array)$/) 
          {$typeString .="double option".$myNo."[128];\n  "; last SWITCH;}
        if($myType =~ /^(float)$/) 
          {$typeString .="float option".$myNo.";\n  "; last SWITCH;}
        }
      }
    }
}
  print $fh <<EOF;
typedef struct $structName
{
 $typeString
}$structName;
EOF
};

sub CreatePrototypeBlock{
my($fh,$myoutFileName,$myDataTypeRef,$charNameRef,$myParamNoRef)=@_;
$myoutFileName =~ s/(\.c)$//;
my $structName = $myoutFileName."Opts";
my $myInputFile = $myoutFileName."InFileName";
my $myOutputFile = $myoutFileName."OutFileName";
$writeOutString="";
$readPrototypeString="";
foreach $myNo (@$myParamNoRef){
    my $key = $charNameRef->[$myNo-1];
    if($key =~ /^-$/) 
    { $writeOutString .="const mxArray* output".$myNo.", "; }
    if($key =~ /^(\+)$/) 
    {
      my $myType = $myDataTypeRef->{$myNo};
    }
}
print $fh <<EOF;

char* $myInputFile=NULL;
char* $myOutputFile=NULL;

void writeOut($writeOutString const char* filename);
$readPrototypeString

void getParameters(int argc, char** argv, $structName* opts);
void printUsage(const char ** argv);
extern int getopt(int argc, char** argv, const char* opts);
extern int opterr;
extern int optind;
extern int optopt;
extern char* optarg;

EOF
};

sub CreateMainFunctionBlock{
my($fh, $mylibName,$myoutFileName,$myDataTypeRef, $rowRef, $colRef, $myParamNoRef,$charNameRef,$descRef) = @_;
$myoutFileName =~ s/(\.c)$//;
my $mylibInit = $mylibName."Initialize()";
my $mylibTerm = $mylibName."Terminate()";
my $structName = $myoutFileName."Opts";

#assemble declaration strings
my  $declareString=" ";
foreach $myNo ( @$myParamNoRef ){
  my $key = $charNameRef->[$myNo-1];
  if($key =~ /^-$/)
  {  $declareString .= "mxArray *output".$myNo." =NULL;\n    ";}
  elsif($key =~ /^(\+)$/)
  {  $declareString .= "char*   inFileName".$myNo." =NULL;\n    ";
     $declareString .= "mxArray *input".$myNo." =NULL;\n    ";
  }
  else
  {  $declareString .= "mxArray *input".$myNo." =NULL;\n    ";}
}

my $getFileNameStrings=" ";
if( length($declareString) > 1)
{
  $declareString .= "char   delimiters[]=\" ,;:\";\n    ";
  $getFileNameStrings="if( ".$myoutFileName."InFileName != NULL) \n   {\n";
  foreach $myNo0 ( @$myParamNoRef ) {
      $key0 = $charNameRef->[$myNo0-1];
      if($key0 =~ /^(\+)$/) 
      {
          if( $myNo0 eq 1 )
          {
          $getFileNameStrings .="    inFileName".$myNo0." = strtok(".$myoutFileName."InFileName,delimiters\);\n";
          }
          else
          {
          $getFileNameStrings .="    inFileName".$myNo0." = strtok(NULL,delimiters);\n";
          }
      }
          }
  $getFileNameStrings .="    };\n";
}

#assemble destroy strings
my  $destroyString="  ";
foreach $myNo ( @$myParamNoRef ){
  my $key = $charNameRef->[$myNo-1];
  if($key =~ /^-$/)
  {  $destroyString .= "mxDestroyArray( output".$myNo." );\n    ";}
  else
  {  $destroyString .= "mxDestroyArray( input".$myNo." );\n    ";}
}

#assemble mxArray initialization strings
my  $mxArrayString="  ";
foreach $myNo (@$myParamNoRef){
    my $createString="";
    $key = $charNameRef->[$myNo-1];
    if($key !~ /^-$/) 
    {
      my $myType = $myDataTypeRef->{$myNo};
      SWITCH: {
        if($myType =~ /^(string)$/) 
          {$createString =" = mxCreateString(opts.option".$myNo; last SWITCH;}
        if($myType =~ /^(int)$/) 
          {$createString =" = mxCreateScalarDouble(opts.option".$myNo; last SWITCH;}
        if($myType =~ /^(double)$/) 
          {$createString =" = mxCreateScalarDouble(opts.option".$myNo; last SWITCH;}
        if($myType =~ /^(array)$/) 
        {
              $rowno=$rowRef->{$myNo};
              $colno=$colRef->{$myNo};
              $createString =" = mxCreateDoubleMatrix(".$rowno.",".$colno.",mxREAL"; last SWITCH;
        }
        if($myType =~ /^(ascii)$/) 
        {
              $rowno=$rowRef->{$myNo};
              $colno=$colRef->{$myNo};
              $createString =" = mxCreateDoubleMatrix(".$rowno.",".$colno.",mxREAL"; last SWITCH;
        }

        if($myType =~ /^(netcdf)$/) 
        {
              $rowno=$rowRef->{$myNo};
              $colno=$colRef->{$myNo};
              $createString =" = mxCreateDoubleMatrix(".$rowno.",".$colno.",mxREAL"; last SWITCH;
        }
        if($myType =~ /^(float)$/) 
          {$createString =" = mxCreateScalarDouble(opts.option".$myNo; last SWITCH;}
      } # end of switch

      $mxArrayString .= "input".$myNo.$createString.");\n  ";
      if($myType =~ /^(array)$/) 
      { 
          $rowno=$rowRef->{$myNo};
          $colno=$colRef->{$myNo};
          $totalNo = $rowno * $colno;
          $mxArrayString
          .="memcpy(mxGetPr(input".$myNo."),opts.option".$myNo.",".$totalNo."*sizeof(double));\n  ";
      }
      if ($myType =~ /^(ascii)$/)
      {
          $rowno=$rowRef->{$myNo};
          $colno=$colRef->{$myNo};
          $totalNo = $rowno * $colno;
          $mxArrayString
          .="readFromAsciiFile(inFileName".$myNo.",mxGetPr(input".$myNo."),".$rowno.",".$colno.");\n";
      }
      if ($myType =~ /^(netcdf)$/)
      {
          $rowno=$rowRef->{$myNo};
          $colno=$colRef->{$myNo};
          $totalNo = $rowno * $colno;
          $varName = $descRef->{$myNo};
          chop($varName);
          $mxArrayString
          .="readFromNetcdfFile(inFileName".$myNo.",mxGetPr(input".$myNo."),\"".$varName."\");\n";
      }
    }
}

my $mlfFunctionString="mlf";
my $programName = $myoutFileName;
$programName =~ s/(\.c)$//;
$programName = ucfirst($programName);
my $listOfparam=" ";
#list output first
my $mytotalOut=0;
foreach $myNo (@$myParamNoRef){
    $key = $charNameRef->[$myNo-1];
    if($key =~ /^-$/) 
    {
        $listOfparam .=",&output".$myNo;
        $mytotalOut++;
    }
}
#list input second
foreach $myNo (@$myParamNoRef){
    $key = $charNameRef->[$myNo-1];
    if($key !~ /^-$/) 
    {
        $listOfparam .=",input".$myNo;
    }
}

$mlfFunctionString .=$programName."(".$mytotalOut.$listOfparam.")";

my  $writeOutString="writeOut(";
foreach $myNo (@$myParamNoRef){
    $key = $charNameRef->[$myNo-1];
    if($key =~ /^-$/) 
    { $writeOutString .="output".$myNo.","; }
}

$writeOutString .=$myoutFileName."OutFileName);";

print $fh <<EOF;
int main(int argc, char** argv)
{
  $declareString

    $structName opts;
    if(argc == 1) { printUsage((const char**) argv); exit(0);}

    /*Call the application initialization routine to make sure that the application
     * was initialized properly.
     */
    if(!mclInitializeApplication(NULL,0) )
    {
        fprintf(stderr,"Could not initialize the application. \\n");
        return -1;
    }

    /*Call the library initialization routine to make sure that the library
     * was initialized properly
     */
     $mylibInit;

    getParameters(argc, argv, &opts);

    $getFileNameStrings
    /*Assign values to parameters */
$mxArrayString

    $mlfFunctionString;

    /*Write out data to a file */
    $writeOutString

    /* Terminate library properly. */
    $mylibTerm;

    /* Release the memory we created */
    $destroyString

    /* Terminate application properly. */
    mclTerminateApplication();
    return 0;
}
EOF
}

sub CreatePrintUsageBlock {
my($fh, $myoutFileName,$myDataTypeRef, $rowRef, $colRef,
$myParamNoRef,$charNameRef,$descRef,$egRef) = @_;

my $commandlineString="printf(\"\%s [-h] [-i <string>] [-o <string>] ";
my $optionsDescString;
foreach $myNo (@$myParamNoRef){
    my $key = $charNameRef->[$myNo-1];
    if($key !~ /^-$/) 
    {
      if($key !~ /^(\+)$/) 
      {
        my $myType = $myDataTypeRef->{$myNo};
        my $desc   = $descRef->{$myNo};
        #print "$desc\n";
        if($myType =~ /^(array)$/) 
        { $commandlineString .= "[-".$key." <number string>] "; }
        else 
        { $commandlineString .= "[-".$key." <".$myType.">] ";} 
        $optionsDescString .="printf(\"-".$key." ".$desc."\\n\");\n"; 
      }
    } #end of if($key !~ /^-$/) 
}
$commandlineString .= "\\n\",argv[0])";

my @egStrings;
my $egDes;
foreach $egDes (@$egRef) {
  push(@egStrings,"printf(\"\%s ".$egDes."\\n\",argv[0]);\n");
}

print $fh <<EOF;
void printUsage(const char ** argv)
{
printf("Usage:\\n");
$commandlineString;
printf("-i Input data file name\\n");
printf("-o Output data file name (Please include directory name. If not, will write to current directory)\\n");
printf("-h Print out this message\\n");
$optionsDescString
printf("Examples:\\n");
@egStrings
}
EOF
}

sub  getArrayStrings{
  my ($myNo,$rowRef,$colRef) = @_;
  my $arrayTemps;
  my @arrayOptions;
  my $arrayStrings;

  push(@arrayOptions,"if(optarg){\n");
  push(@arrayOptions,"tmp=malloc(strlen(optarg)*sizeof(char));\n");
  push(@arrayOptions,"strcpy(tmp,optarg);\n");

  my $row = $rowRef->{$myNo};
  my $col = $colRef->{$myNo};
  my $total = $row * $col;
  for($i=0;$i<$total;$i++)
  {
      $arrayTemps .="  char* tmp".$i." =NULL;/* Temperory pointer */\n";
      if ( $i != 0 ) # skip first one
      {
        push(@arrayOptions,"tmp".$i." = strtok(NULL, \" ,;:\");\n");
        push(@arrayOptions,"if(tmp".$i." != NULL) \n");
        push(@arrayOptions,"{opts->option".$myNo."[".$i."]"."=atof(tmp".$i.");}\n");
        push(@arrayOptions,"else{ opts->option".$myNo."[".$i."]"."=0;};\n");
      }
      else # first one
      {
        push(@arrayOptions,"tmp".$i." = strtok(tmp, \" ,;:\");\n");
        push(@arrayOptions,"if( tmp".$i." != NULL) \n");
        push(@arrayOptions,"{opts->option".$myNo."[0]"."=atof("."tmp".$i.");}\n");
        push(@arrayOptions,"else{ opts->option".$myNo."[0]"."=0;};\n");
      }

  } # end of for loop
  
  push(@arrayOptions,"}\n");
  foreach $myline (@arrayOptions)
  {$arrayStrings .= $myline; }
  return ($arrayTemps, $arrayStrings);
}

sub CreateGetCommandlineOptionBlock
{
my($fh, $myoutFileName,$myDataTypeRef, $rowRef, $colRef, $myParamNoRef,$charNameRef,$descRef) = @_;
$myoutFileName =~ s/(\.c)$//;
my $myInputFile = $myoutFileName."InFileName";
my $myOutputFile = $myoutFileName."OutFileName";
my $optionsString;
foreach $myNo (@$myParamNoRef){
    my $key = $charNameRef->[$myNo-1];
    if($key !~ /^-$/) 
    {
      if($key !~ /^(\+)$/) 
      {
        $optionsString .=$key.":";
        #print "$desc\n";
      }
   }
}
$optionsString .="i:o:h";

my @caseStrings;
my $temps="";

foreach $myNo (@$myParamNoRef){
  my $key = $charNameRef->[$myNo-1];
  if( $key !~ /^(-)$/) 
  {
    if($key !~ /^(\+)$/) 
    {
      $caseStrings[$myNo-1]="    case '";
      my $convertString="";
      my $elseString="\n";
      my $myType = $myDataTypeRef->{$myNo};
      SWITCH: {
      if($myType =~ /^(string)$/) 
        {$convertString ="strcpy(opts->option".$myNo.","; last SWITCH;}
      if($myType =~ /^(int)$/) 
        {$convertString .="atoi("; $elseString="        else
        opts->option".$myNo."=0;\n";last SWITCH;}
      if($myType =~ /^(double)$/) 
        {$convertString .="atof("; $elseString="        else
        opts->option".$myNo."=0;\n";last SWITCH;}
      if($myType =~ /^(array)$/) 
        { ($temps,$convertString) = getArrayStrings($myNo,$rowRef,$colRef);last SWITCH; }
      if($myType =~ /^(float)$/) 
        {$convertString .="atof(";$elseString="        else
        opts->option".$myNo."=0;\n";last SWITCH;}
      }
      if($myType =~ /^(string)$/) 
      { $caseStrings[$myNo-1] .=$key."':\n        if(optarg)\n     ".$convertString."optarg);\n".$elseString."      break;\n";}
      elsif( $myType =~ /^(array)$/) 
      { $caseStrings[$myNo-1] .=$key."':\n        ".$convertString."\n      break;\n";}
      else
      {$caseStrings[$myNo-1] .=$key."':\n        if(optarg)\n      opts->option".$myNo."=".$convertString."optarg);\n".$elseString." break;\n";}
    }
  }
}
$optionsString .="i:o:h";

$myoutFileName =~ s/(\.c)$//;
my $structName = $myoutFileName."Opts";

print $fh <<EOF;
void getParameters(int argc, char** argv, $structName * opts)
{
  int c; /* keep option char */
  char *tmp=NULL;
$temps

  if(argc == 1) return; /* return if no options appear */
                                                                                                 
  /* Get the options */
  while ( c = getopt(argc, (char**) argv, "$optionsString") ) { 
    if (c == -1 ) break;

    switch (c) {
    case 'i':
        if (optarg)
        $myInputFile=optarg;
        break;

    case 'o':
        if (optarg)
        $myOutputFile=optarg;
        break;

    case 'h':
        printUsage((const char**) argv);
        exit(0);

@caseStrings
    default:
        printf ("?? getopt returned input character code '\%o' ??\\n", c);
    } /* end of swith */
  }
}
EOF
}
sub CreateWriteOutDataBlock
{
my ($fh,$myoutFileName,$dataTypesRef,$rowNumsRef,$colNumsRef,$paramNosRef,$charNamesRef) = @_;
my $writeOutString="";

$declareString = "char   delimiters[]=\" ,;:\";\n    ";
$getFileNameStrings="if( filename != NULL) \n   {\n";
$lastParam;
foreach $myNo (@$paramNosRef){
    my $key = $charNamesRef->[$myNo-1];
    if($key =~ /^-$/) 
    {
      $lastParam = $myNo -1; last;
    }
}
foreach $myNo (@$paramNosRef){
    my $key = $charNamesRef->[$myNo-1];
    if($key =~ /^-$/) 
    { 
        $writeOutString .="const mxArray* output".$myNo.", "; 
        $declareString .="char * outFileName".$myNo.";\n";
          if( ($myNo - $lastParam) eq 1 )
          {
          $getFileNameStrings .="    outFileName".$myNo." = strtok("."(char*) filename,delimiters\);\n";
          }
          else
          {
          $getFileNameStrings .="    outFileName".$myNo." = strtok(NULL,delimiters);\n";
          }
    }
}
$getFileNameStrings .="    };\n";

print $fh <<EOF;

/* This function will writeOut the double matrix stored in an mxArray.
 * This function assumes that the mxArray passed as input contains double
 * array.
 */
void writeOut($writeOutString const char *filename)
{
    int i=0, j=0; /* loop index variables */
    int r=0, c=0; /* variables to store the row and column length of the matrix */
    double *data=NULL; /* variable to point to the double data stored within the mxArray */
    const int * dims=NULL;
    FILE* fp=NULL;
    $declareString
    $getFileNameStrings

    /* Get the size of the matrix and a pointer to mxArray of data*/
EOF

foreach $myNo (@$paramNosRef){
    my $key = $charNamesRef->[$myNo-1];
    if($key =~ /^-$/) 
    { 
print $fh <<EOF;
    fp=fopen(outFileName$myNo,"w");
    r = mxGetM(output$myNo); /* total row number */
    c = mxGetN(output$myNo); /* total columns number */
    data = mxGetPr(output$myNo);
    
    /* Loop through the data and write data to output file */
    for( j = 0; j < r; j++)
    {
        for( i = 0; i < c; i++ ){
            fprintf(fp,"\%15.7e ",data[j+i*r]); 
    /*FIXME: Matlab mcc has bug that row and column is swapped inside call.*/
        }
        fprintf(fp,"\\n");
    }
    fclose(fp);
EOF
    }
}
print $fh <<EOF;
}
EOF
}
