#include <stdio.h>
#include "netcdfRW.h"

void printError(int status) {
  if (status != NC_NOERR) {
    fprintf(stderr, "%s\n", nc_strerror(status));
    exit(-1);
  }
}

void readFromNetcdfFile(const char* filename,double * fillMe, char* varName)
{
    fprintf(stderr,"Reading the netcdf file %s... ",filename);
    if(filename == NULL) 
    {
        fprintf(stderr,"filename is null...exit\n");
        exit(0) ;
    }
    if(fillMe == NULL) 
    {
        fprintf(stderr,"fillMe is null...exit\n");
        exit(0) ; 
    }

    int ncid;
    int status = nc_open(filename, NC_NOWRITE, &ncid);
    printError(status);

    int dataid;
    status = nc_inq_varid(ncid,varName,&dataid);
    printError(status);

    nc_type data_type;
    int data_ndims;
    int data_dims[NC_MAX_VAR_DIMS];
    int data_natts;
    status = nc_inq_var(ncid, dataid, 0, &data_type,&data_ndims,data_dims,&data_natts);
    printError(status);

    size_t totalNum=1;
    int i=0;
    int j=0;
    for(i=0;i<data_ndims;i++)
    {
      size_t size;
      status = nc_inq_dimlen(ncid,data_dims[i],&size);
      printError(status);
      totalNum *=size;
    }
    if(data_type == NC_FLOAT)
    {
      float *fp = (float*) calloc( totalNum, sizeof(float));
      status = nc_get_var_float(ncid,dataid,fp);
      printError(status);
      int rowNum, colNum;
      status = nc_inq_dimlen(ncid,data_dims[0],&rowNum);
      status = nc_inq_dimlen(ncid,data_dims[1],&colNum);
      for(i=0; i<rowNum; i++)
      for(j=0; j<colNum; j++)
      { 
          fillMe[j*rowNum+i] = fp[i*colNum+j];
      }
      free(fp);
    }
        
    if(data_type == NC_DOUBLE)
    {
      double *dp = (double*) calloc(totalNum, sizeof(double));
      status = nc_get_var_double(ncid,dataid,dp);
      printError(status);

      int rowNum, colNum;
      status = nc_inq_dimlen(ncid,data_dims[0],&rowNum);
      status = nc_inq_dimlen(ncid,data_dims[1],&colNum);
      for(i=0; i<rowNum; i++)
      for(j=0; j<colNum; j++)
      { 
          fillMe[j*rowNum+i] = dp[i*colNum+j];
      }
      free(dp);
    }
    nc_close(ncid);
    fprintf(stderr,"finished.\n ",filename);
}
